/* include/isl-interface/config.h.  Generated from config.h.in by configure.  */
#ifndef ISL_CLANG_WRAP_CONFIG_H
#define ISL_CLANG_WRAP_CONFIG_H

#define CREATE_FROM_ARGS_TAKES_ARRAYREF /**/
#define CREATEDIAGNOSTICS_TAKES_VFS /**/
/* #undef CREATETARGETINFO_TAKES_SHARED_PTR */
#define HAVE_LLVM_OPTION_ARG_H /**/
#define HAVE_TARGETPARSER_HOST_H /**/
#define IK_C Language::C
#define ISL_CLANG_PREFIX "/usr/lib/llvm-21"
#define ISL_CLANG_RESOURCE_DIR "/usr/lib/llvm-21/lib/clang/21"
/* #undef ISL_CLANG_SYSROOT */
#define SETLANGDEFAULTS LangOptions
#define SETLANGDEFAULTS_TAKES_5_ARGUMENTS /**/

#endif
